/***************************************************************************
                          prototypeview.cpp  -  description
                             -------------------
    begin                : Thu May 15 22:50:15 CEST 2003
    copyright            : (C) 2003 by Marcus Beranek
    email                : marcus@beranek.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

//
// This class creates the main widget, which displays all information
// about our protocol. On the left we have the listing of all protocol-
// files. At the top, we have all blocks of a selected protocol and
// underneth the blocks there is the description of a selected block.
//

#include <qvariant.h>
#include <qframe.h>
#include <qlabel.h>
#include <qlistbox.h>
#include <qpushbutton.h>
#include <qsplitter.h>
#include <qtextedit.h>
#include <qlayout.h>
#include <qtooltip.h>
#include <qwhatsthis.h>
#include <qaction.h>
#include <qmenubar.h>
#include <qpopupmenu.h>
#include <qtoolbar.h>
#include <qhbox.h>
#include <qcolor.h>
#include <qwhatsthis.h>
#include <qregexp.h>
#include <qmainwindow.h>
#include <qptrlist.h>
#include <qmessagebox.h>

// #include <qcanvas.h>
// #include <qpainter.h>
// #include <qpen.h>

#include <dlfcn.h>


#include <iostream>
#include <iomanip>

#include "prototypeview.h"
#include "block.h"
#include "blocklist.h"
#include "blockbtn.h"
// #include "treedisplay.h"

ProtoTypeView::ProtoTypeView(QWidget *parent, ProtoTypeDoc *doc) : QWidget(parent)
{
	/** connect doc with the view*/
	connect(doc, SIGNAL(documentChanged()), this, SLOT(slotDocumentChanged()));

	buttonlist.setAutoDelete( true );
	buttonlist2.setAutoDelete( true );


	ProtoTypeLayout = new QHBoxLayout( this, 11, 6, "ProtoTypeLayout");
	splitter1 = new QSplitter( this, "splitter1" );
	splitter1->setOrientation( QSplitter::Horizontal );

	listBox2 = new QListBox( splitter1, "listBox2" );
	listBox2->setSizePolicy( QSizePolicy( (QSizePolicy::SizeType)5, (QSizePolicy::SizeType)7, 0, 0, listBox2->sizePolicy().hasHeightForWidth() ) );
	listBox2->setMinimumSize( QSize( 30, 100 ) );

	QWhatsThis::add( listBox2, tr( "Listing of all known protocols." ) );

	frame4 = new QFrame( splitter1, "frame4" );
	frame4->setSizePolicy( QSizePolicy( (QSizePolicy::SizeType)7, (QSizePolicy::SizeType)7, 0, 0, frame4->sizePolicy().hasHeightForWidth() ) );
	frame4->setMinimumSize( QSize( 350, 400 ) );
	frame4->setFrameShape( QFrame::StyledPanel );
	frame4->setFrameShadow( QFrame::Raised );
	frame4Layout = new QVBoxLayout( frame4, 11, 6, "frame4Layout");

	splitter3 = new QSplitter( frame4, "splitter3" );
	splitter3->setOrientation( QSplitter::Vertical );

	frame5 = new QScrollView( splitter3, "frame5" );
	frame5->setMinimumSize( QSize( 300, 50 ) );
	frame5->setFrameShape( QFrame::StyledPanel );
	frame5->setFrameShadow( QFrame::Raised );
	frame5->setSizePolicy( QSizePolicy::Expanding, QSizePolicy::Minimum );
	frame5->setGeometry( QRect( 0, 0, 300, 50 ) );
	frame5->setHScrollBarMode( QScrollView::AlwaysOn );

	QWhatsThis::add( frame5, tr( "Listing of the protocol-blocks." ) );

	blockbox = new QHBox( frame5->viewport());
	QWhatsThis::add( blockbox, tr( "Listing of the protocol-blocks." ) );
	frame5->addChild(blockbox);

	frame6 = new QFrame( splitter3, "frame6" );
	frame6->setFrameShape( QFrame::StyledPanel );
	frame6->setFrameShadow( QFrame::Raised );
	frame6Layout = new QVBoxLayout( frame6, 11, 6, "frame6Layout");

	infobox = new QHBox( frame6 );
	frame6Layout->addWidget( infobox );

	textLabel1 = new QLabel( infobox, "blocknamelabel" );
	textLabel1->setMinimumSize( QSize( 100, 20 ) );
	textLabel1->setFrameStyle( QFrame::Box | QFrame::Raised );
	textLabel1->setLineWidth(1);

	QWhatsThis::add( textLabel1, tr( "Blockname" ) );

	textLabel2 = new QLabel( infobox, "blockfunctionlabel" );
	textLabel2->setMinimumSize( QSize( 100, 20 ) );
	textLabel2->setFrameStyle( QFrame::Box | QFrame::Raised );
	textLabel2->setLineWidth(1);

	QWhatsThis::add( textLabel2, tr( "Blockfunction" ) );

	textLabel3 = new QLabel( infobox, "blocktypelabel" );
	textLabel3->setMinimumSize( QSize( 100, 20 ) );
	textLabel3->setFrameStyle( QFrame::Box | QFrame::Raised );
	textLabel3->setLineWidth(1);

	QWhatsThis::add( textLabel3, tr( "Blocktype" ) );

	textLabel4 = new QLabel( infobox, "blockstate" );
	textLabel4->setMinimumSize( QSize( 100, 20 ) );
	textLabel4->setFrameStyle( QFrame::Box | QFrame::Raised );
	textLabel4->setLineWidth(1);

	QWhatsThis::add( textLabel4, tr( "Blockstate" ) );

	textEdit1 = new QTextEdit( frame6, "blockdescription" );
	textEdit1->setMinimumSize( QSize( 300, 250 ) );
	QWhatsThis::add( textEdit1, tr( "Description of the selected block." ) );

	frame6Layout->addWidget( textEdit1 );

	sequencebox = new QHBox( frame6 );
	sequencebox->setMinimumSize( QSize( 300, 150 ) );
	frame6Layout->addWidget( sequencebox );

	followslabel = new QLabel( sequencebox, "blockollows" );
	followslabel->setMinimumSize( QSize( 150, 150 ) );
	followslabel->setFrameStyle( QFrame::Box | QFrame::Raised );
	followslabel->setLineWidth(1);
	followslabel->setAlignment( Qt::AlignAuto | Qt::AlignTop );

	followedbylabel = new QLabel( sequencebox, "blockollows" );
	followedbylabel->setMinimumSize( QSize( 150, 150 ) );
	followedbylabel->setFrameStyle( QFrame::Box | QFrame::Raised );
	followedbylabel->setLineWidth(1);
	followedbylabel->setAlignment( Qt::AlignAuto | Qt::AlignTop );

	frame4Layout->addWidget( splitter3 );
	ProtoTypeLayout->addWidget( splitter1 );

	resize( QSize(800, 600).expandedTo(minimumSizeHint()) );

	// signals and slots connections
	connect( listBox2, SIGNAL( clicked(QListBoxItem*) ), this, SLOT( loadBlocks(QListBoxItem*) ) );

	state = ShowAll;
	currentBlockName = "";
	currentProtocol = "";

	// weiteres Fenster bereitstellen, aber noch nicht zeigen:

	max_button_width = 0;
	max_button_height = 0;
}
//---------------------------------------------------------------------------


ProtoTypeView::~ProtoTypeView()
{
}
//---------------------------------------------------------------------------


void ProtoTypeView::slotDocumentChanged()
{
	// Okay: at this point, the doc has built up its protocol-tree.
	// Now we look for the blocklist and construct the buttons for each block:

	buttonlist.clear();
	buttonlist2.clear();

	max_button_width = 0;

	textLabel1->setText("");
	textLabel2->setText("");
	textLabel3->setText("");
	textLabel4->setText("");
	textEdit1->setText("");

	BlockList* bl = doc->getBlockList();
	currentProtocol = doc->getCurrentProtocol();
	garbageCollect(); // kann ja nicht schaden... ;)

	if(! bl->isEmpty() ){
		parseTree( bl, bl->first(), state );
		doc->clearVisited();
		doc->clearPrinted();
		treeview_x_coord = 0;
		treeview_y_coord = 0;
		col=0;
		row=0;
		newID = 1;

		if( ! pluglist.isEmpty() ){
			for( ViewPlugin* vp = pluglist.first(); vp; vp = pluglist.next() ){
				initPlugin( vp );
			}
		}

		doc->clearVisited();
		doc->clearPrinted();


	}

}
//---------------------------------------------------------------------------


void ProtoTypeView::parseTree( BlockList* bl, Block* b, ShowState show )
{
	//
	// This method prints a button for every block in the protocol-tree.
	// The easiest way to parse the tree would be a deep-search.
	// Unfortunatly the buttons would be in the wrong order. So we are
	// parsing the tree more like a broad-search. The only problem is,
	// that this implementation makes heavy use of the "find"-method,
	// which - in the worst case - has to go through the whole list!
	//

	Block* block;
	BlockBtn* bb;

	// Search for the block b ...
	int z = bl->find( b );
	if ( z > -1 ){
		block = bl->at( z );
		block->visited = true;

		// ... then get list of all following blocks ...
		SuccessorList sl;
		switch( show ){
		case ShowAll: {
				sl = block->getFollowedby();
				break;
			}
		case ShowTypical: {
				sl = block->getTypicalFollowedby();
				break;
			}
		}
		Successor* f;

		// ... print evey successor (we do not print the root-node,
		//     cause it contains the protocol-name only) ...
		for( f=sl.first(); f; f=sl.next() ){
			int k = bl->find( new Block( f->ID ) );
			if( k > -1 ){
				Block* v = bl->at( k );
				if( ! v->printed ){
					v->printed=true;
					bb = new BlockBtn( blockbox );
					bb->setID( v->ID );
					bb->setText( v->Command );
					bb->setPaletteBackgroundColor( colorChooser( v->Depth ) );
					bb->setPaletteForegroundColor( QColor( 0, 0, 0 ) );
					QWhatsThis::add( bb, tr( "Protocol-Block." ) );
					buttonlist.append( bb );
					bb->show();

					int w_tmp = v->Command.length() * 12 + 20 ;

					max_button_width = ( w_tmp > max_button_width ) ? ( w_tmp ) : ( max_button_width );
					max_button_height = ( bb->height() > max_button_height ) ? ( bb->height() ) : ( max_button_height );

					connect( bb, SIGNAL( blockSelected(int) ), this, SLOT( showBlockDescription(int) ) );
				}
			}
		}

		for( f=sl.first(); f; f=sl.next() ){
			int r = bl->find( new Block( f->ID ) );
			if ( r > -1 ){
				Block* v = bl->at( r );
				// .. decent recursive into subtree, if we yet have not been there ...
				if(! v->visited ){
					parseTree( bl, new Block( f->ID ), show );
				}
			}
		}
	}
}
//---------------------------------------------------------------------------


void ProtoTypeView::launchPlugin( QString pl )
{
	garbageCollect();
	mw = new QMainWindow;
	winlist.append( mw );
	treeLayout = new QHBoxLayout( mw, 11, 6, "treeLayout");

	mw->setGeometry(20, 20, 800, 600 );
	mw->hide();

	//plugindisplay = loadPlugin( pl , tree->viewport() );
	plugindisplay = loadPlugin( pl , mw );
	if( plugindisplay ){
		std::cout << "Plugin " + pl + " created" << endl;
		initPlugin( plugindisplay );
		treeLayout->addWidget( plugindisplay );
		mw->setCentralWidget( plugindisplay );
		mw->show();
	}
	
}
//---------------------------------------------------------------------------


void ProtoTypeView::garbageCollect()
{
	// Remove all plugins, which are no longer displayed...
	//
	//
	// TODO: fix following problem:
	// open plugin1, open plugin2, plugin3, close plugin1, open plugin4 => plugin2 gets deleted... :(
	// (unfortunatly not exactly this way, I have to try out a little bit more to reproduce this error)
	//
	//
	
	int wCount = 0;
	int pCount = 0;
	
	if( ! winlist.isEmpty() ){
		wCount = winlist.count();
	}
	if( ! pluglist.isEmpty() ){
		pCount = pluglist.count();
	}
	
	for( int i = 0; i< wCount; i++ ){
		QMainWindow* m = winlist.at(i);
		ViewPlugin*  v = pluglist.at(i);
		
		if( ! m->isVisible() ){
			
			// A poor plugin without window... :(
			
			pluglist.remove( i );
			if( v ){
				std::cout << "removing plugin " << v << endl;
				destroy_plug( v );
				pCount--;
				
				winlist.remove(i);
				if( m ){
					delete m;
					wCount--;
				}
			}
			
		}
		
	}
	
}
//---------------------------------------------------------------------------


void ProtoTypeView::loadBlocks( QListBoxItem* item)
{
	if(item){
		emit( protocolSelected( item->text() + ".xml" ) );
	}
}
//---------------------------------------------------------------------------


void ProtoTypeView::setShowState( ShowState s ){
	state = s;
	slotDocumentChanged();
}
//---------------------------------------------------------------------------


void ProtoTypeView::addProtocol( QString p )
{
	// TODO: Im Listing sind im Augenblick nur die Dateinamen abgelegt.
	// Hier sollten zur besseren bersicht die Protokollnamen stehen.
	// Es muss also noch eine Liste ProtokollName <-> Dateiname angelegt werden.
	if( int pos = p.find( QRegExp("\\.xml"), 0) ){
		p.truncate( pos );
		listBox2->insertItem( p );
	}

}
//---------------------------------------------------------------------------


void ProtoTypeView::clearProtocolList()
{
	listBox2->clear();
	garbageCollect(); // kann ja nicht schaden... ;)
}
//---------------------------------------------------------------------------


QColor  ProtoTypeView::colorChooser(int c)
{
	QColor color[10];

	color[0] = QColor( 0xFF, 0xA8, 0xA9 );
	color[1] = QColor( 0xFC, 0xFF, 0xE0 );
	color[2] = QColor( 0xAA, 0xD3, 0xFF );
	color[3] = QColor( 0xEB, 0xF6, 0xFF );
	color[4] = QColor( 0xD9, 0xFF, 0xB0 );
	color[5] = QColor( 0xE3, 0xFF, 0xEB );
	color[6] = QColor( 0xA5, 0xFF, 0xAE );
	color[7] = QColor( 0xFF, 0xDE, 0xB2 );
	color[8] = QColor( 0xD8, 0xDA, 0xFF );
	color[9] = QColor( 0xFF, 0xF9, 0xEE );

	int i = c % 10;

	return color[i];
}
//---------------------------------------------------------------------------


void ProtoTypeView::setDoc( ProtoTypeDoc* d )
{
	doc = d;
}
//---------------------------------------------------------------------------


void ProtoTypeView::showBlockDescription( int id )
{

	BlockList* bl = doc->getBlockList();
	Block* b;

	int i = bl->find( new Block( id ) );
	if ( i > -1 ){
		b = bl->at(i);

		currentBlockName = b->Name;

		QString d;
		d.setNum( b->Depth );

		textEdit1->setText( b->Description );
		textLabel1->setText( b->Name );
		textLabel2->setText( b->Function );
		textLabel4->setText( tr("State: ") + b->State );

		SuccessorList sl = b->getFollowedby();
		Successor* s;
		QString f1 = tr( "Followed by:\n" );
		for( s = sl.first(); s; s = sl.next() ){
			f1.append( doc->getBlock( s->ID )->Command + "\n" );
		}
		followslabel->setText( f1 );

		sl = b->getFollows();
		QString f2 = tr( "Follows:\n" );
		for( s = sl.first(); s; s = sl.next() ){
			f2.append( doc->getBlock( s->ID )->Command + "\n" );
		}
		followedbylabel->setText( f2 );

		switch( b->Type ){
		case Block::Send : {
				textLabel3->setText( tr("SEND") );
				break;
			}
		case Block::Receive : {
				textLabel3->setText( tr("RECEIVE") );
				break;
			}
		default : {
				textLabel3->setText( tr("-") );
				break;
			}

		}

	}
	//std::cout << "blockbutton " << id << " was clicked." << std::endl;
}
//---------------------------------------------------------------------------

QTextEdit* ProtoTypeView::getTextEdit()
{
	return textEdit1;
}
//---------------------------------------------------------------------------


QString ProtoTypeView::getCurrentBlockName()
{
	return currentBlockName;
}
//---------------------------------------------------------------------------


QString ProtoTypeView::getCurrentProtocol()
{
	return currentProtocol;
}
//---------------------------------------------------------------------------


void ProtoTypeView::setCurrentProtocol( QString p )
{
	currentProtocol = p;
}
//---------------------------------------------------------------------------


void ProtoTypeView::setTreeView( bool on )
{
	if( on ){
		// parse tree, build flowchart...
		mw->show();

	} else {
		mw->hide();

	}
}
//---------------------------------------------------------------------------


void ProtoTypeView::unloadPlugin( ViewPlugin* plug )
{
	if( plug ){
		tree->removeChild( plug );
		destroy_plug( plug );
	}
}

ViewPlugin* ProtoTypeView::loadPlugin( QString pluginlib, QWidget* par )
{
	ViewPlugin* v = NULL;
	
	// now: load pluginlibrary...
	//backendlib = dlopen("libprototypetreeview.so", RTLD_LAZY);

	backendlib = dlopen( pluginlib.latin1(), RTLD_LAZY);
	
	if (!backendlib){
		// hmm... no lib... :|
		QString err = tr("Error loading ") + pluginlib + "\n" + dlerror();
		//std::cerr << err << dlerror() << endl;
		QMessageBox* mb = new QMessageBox( this );
		mb->setText( err );
		mb->setIcon( QMessageBox::Critical );
		mb->show();
	} else {
		create_plug  = (create_t*)  dlsym( backendlib, "create" );
		destroy_plug = (destroy_t*) dlsym( backendlib, "destroy" );
		
		if ( !create_plug || !destroy_plug ){
			std::cerr << dlerror() << endl;
		} else {
			v = create_plug( par );
			pluglist.append( v );
		}
	}
	return v;
}
//---------------------------------------------------------------------------

void ProtoTypeView::initPlugin( ViewPlugin* plugin )
{
	if (plugin){
		plugin->clear();
		
		Block* f;
		
		//
		// TODO: initialize plugin just by passing one block-list (or something similar...)
		//
		
		BlockList* bl = doc->getBlockList(); 
		
		for( f=bl->first(); f; f=bl->next() ){

			plugin->addBlock( f );
			// Da es im Augenblick noch Probeme mit der QPtrList gibt (diese wird nicht bergeben),
			// mssen die Listen in einem Block alle separat angelegt werden... 
			SuccessorList sl;
			
			sl = f->getFollowedby();
			if(!sl.isEmpty()){
				for( Successor* sux = sl.first(); sux; sux = sl.next() ){
					//std::cout << "correcting block " << f->ID << " follower " << sux->ID << endl;
					plugin->addBlockFollowedby( f->ID, sux->ID );
				}
			}
			sl = f->getFollows();
			if(!sl.isEmpty()){
				for( Successor* sux = sl.first(); sux; sux = sl.next() ){
					//std::cout << "correcting block " << f->ID << " follower " << sux->ID << endl;
					plugin->addBlockFollows( f->ID, sux->ID );
				}
			}
			sl = f->getTypicalFollowedby();
			if(!sl.isEmpty()){
				for( Successor* sux = sl.first(); sux; sux = sl.next() ){
					//std::cout << "correcting block " << f->ID << " follower " << sux->ID << endl;
					plugin->addBlockTypicalFollowedby( f->ID, sux->ID );
				}
			}
			
		}
		plugin->init( (ViewPlugin::ShowState) state );
	}
}
//---------------------------------------------------------------------------


